#include <ESP8266_Lib.h>
#include <BlynkSimpleShieldEsp8266.h>
#include <Adafruit_BMP280.h>
#include <SoftwareSerial.h>
#include <LiquidCrystal.h>
#include "RTClib.h"
#include <Wire.h>

#define ESP8266_BAUD 9600
#define BLYNK_PRINT Serial

#define currentMonth 6

/* RX, TX and LCD pins */
#define pin_RX 2
#define pin_TX 3
#define pin_RS 8
#define pin_EN 9
#define pin_D4 4
#define pin_D5 5
#define pin_D6 6
#define pin_D7 7

/* Sea Level Pressure (hPa) and current location altitude */
#define SLP_HPA (1013.25)
#define ALTITUDE_HOME (131)

/* Thingspeak constants */
#define tempField "field1"
#define pressField "field2"
#define PORT "80"

char auth[] = "aDW9BdHPdYZSCskBZ2bQ-9sqb-JUTu5N";
char ssid[] = "DIGI-k6Ye";
char pass[] = "J49pkqZN";

LiquidCrystal lcd(pin_RS, pin_EN, pin_D4, pin_D5, pin_D6, pin_D7);
SoftwareSerial EspSerial(pin_RX, pin_TX);
ESP8266 wifi(&EspSerial);
Adafruit_BMP280 bmp;
RTC_DS1307 RTC;

String API_key = "8DIF4N8PD7ID76O1";
String HOST = "api.thingspeak.com";

int Z = 0;
int commandRetries;
int loopNumber = 0;
int canCalculateZambrettiCounter = 0;
int pressureEntriesCount = 1;
boolean success = false;
float pressureArray[10] = {0};
String interpretedWeather = "Wait";

void setup()
{
  Serial.begin(9600);
  lcd.begin(16, 2);

  // Set ESP8266 baud rate
  EspSerial.begin(ESP8266_BAUD);

  EspSerial.println("AT+UART_DEF=9600, 8, 1, 0, 3");
  delay(10);
  Blynk.begin(auth, wifi, ssid, pass);

  if (!bmp.begin()) {
    Serial.println(F("Could not find a valid BMP280 sensor, check wiring or "
                     "try a different address!"));
    while (1) delay(10);
  }

  bmp.setSampling(Adafruit_BMP280::MODE_NORMAL,     /* Operating Mode. */
                  Adafruit_BMP280::SAMPLING_X2,     /* Temp. oversampling */
                  Adafruit_BMP280::SAMPLING_X16,    /* Pressure oversampling */
                  Adafruit_BMP280::FILTER_X16,      /* Filtering. */
                  Adafruit_BMP280::STANDBY_MS_500); /* Standby time. */
}

void loop()
{
  //Blynk.run();
  float pressure = bmp.readPressure() / 100.0F;
  float temperature = bmp.readTemperature();
  float seaPressure = stationToSeaLevel(pressure, ALTITUDE_HOME, temperature);

  lcd.clear();
  printPressureLCD(seaPressure);
  printTemperatureLCD(temperature);
  printPredictionLCD();

  /* Send data to Thingspeak (delay 250) */
  sendData(temperature, pressure);

  /* Simulate the passing of one second (cumulates with the 250ms delay from before) */
  delay(750);

  /* The number of "seconds" past */
  loopNumber++;

  /* After every 60 seconds, add a new entry in the pressure array
     Shift the entries to the left and remove the first one when adding a new entry */
  if (loopNumber == 60) {
    loopNumber = 0;

    if (pressureEntriesCount == 10) {
      for (int i = 0; i < 9; i++) {
        pressureArray[i] = pressureArray[i + 1];
      }

      pressureArray[pressureEntriesCount - 1] = seaPressure;
    } else {
      pressureArray[pressureEntriesCount] = seaPressure;
      pressureEntriesCount++;
    }
  }

  /* Wait until 10 minutes have passed to apply Zambretti */
  if (canCalculateZambrettiCounter < 600) {
    canCalculateZambrettiCounter++;
    return;
  }

  /* Calculate Zambretti and interpret the weather based on the Z variable returned
     and the current month */
  float lastEntriesAverage = (pressureArray[9] + pressureArray[8] + pressureArray[7]) / 3;
  float firstEntriesAverage = (pressureArray[0] + pressureArray[1] + pressureArray[2]) / 3;

  Z = calcZambretti(lastEntriesAverage, firstEntriesAverage, currentMonth);
  interpretedWeather = interpretWeather(Z, seaPressure);
}

void printPressureLCD(float pressure) {
  String presString = String(pressure, 0);
  lcd.setCursor(0, 1);
  lcd.print("P:");
  lcd.print(presString);
  lcd.print("hPa");
}

void printTemperatureLCD(float temperature) {
  String tempString = String(temperature, 0);
  lcd.setCursor(0, 0);
  lcd.print("T:");
  lcd.print(tempString);
  lcd.print((char) 223);
  lcd.print("C");
}

void printPredictionLCD() {
  lcd.setCursor(7, 0);
  lcd.print(interpretedWeather);
  lcd.setCursor(12, 1);
  lcd.print("Z=");
  lcd.print(Z);
}

/* Convert current pressure to sea level pressure */
float stationToSeaLevel(float p, int height, float t) {
  return (double) p * pow(1 - 0.0065 * (double) height / (t + 0.0065 * (double) height + 273.15), -5.275);
}

/* Send data to Thingspeak */
void sendData(float temperature, float pressure) {
  /* Print temperature and pressure values to the console */
  Serial.println("Pressure: " + String(pressure, 2) + "        Temp: " + String(temperature, 1));

  /* Build GET request URL */
  String getData = "GET /update?api_key=" + API_key + "&" + tempField + "=" + String(temperature, 2) + "&" + pressField + "=" +
                   String(pressure, 2);

  /* Set a single connection */
  sendCommand("AT+CIPMUX=1", 1, "OK");

  /* Initiate the TCP connection on port 80 to the Thingspeak API */
  sendCommand("AT+CIPSTART=0,\"TCP\",\"" + HOST + "\"," + PORT, 1, "OK");

  /* Prepare to send data */
  sendCommand("AT+CIPSEND=0," + String(getData.length() + 4), 1, ">");

  /* Send data to the ESP */
  EspSerial.println(getData);

  delay(250);

  /* Close TCP connection */
  sendCommand("AT+CIPCLOSE=0", 1  , "OK");
}

/* Send commands to the ESP */
void sendCommand(String command, int maxRetries, char expectedReply[]) {
  Serial.print(command + " ");

  /* Send the command to ESP up to a maximum number of retries */
  while (commandRetries < maxRetries) {
    EspSerial.println(command);
    if (EspSerial.find(expectedReply)) {
      success = true;
      break;
    }

    commandRetries++;
  }

  if (success) {
    Serial.println("OK");
  } else {
    Serial.println("Fail");
  }

  commandRetries = 0;
  success = false;
}

int calcZambretti(float currPressure, float prevPressure, int mon) {
  if (currPressure < prevPressure) {
    /* FALLING */
    if (mon >= 4 and mon <= 9) {
      /* SUMMER */
      if (currPressure >= 1030)
        return 2;
      else if (currPressure >= 1020 and currPressure < 1030)
        return 8;
      else if (currPressure >= 1010 and currPressure < 1020)
        return 18;
      else if (currPressure >= 1000 and currPressure < 1010)
        return 21;
      else if (currPressure >= 990 and currPressure < 1000)
        return 24;
      else if (currPressure >= 980 and currPressure < 990)
        return 24;
      else if (currPressure >= 970 and currPressure < 980)
        return 26;
      else if (currPressure < 970)
        return 26;
    } else {
      /* WINTER */
      if (currPressure >= 1030)
        return 2;
      else if (currPressure >= 1020 and currPressure < 1030)
        return 8;
      else if (currPressure >= 1010 and currPressure < 1020)
        return 15;
      else if (currPressure >= 1000 and currPressure < 1010)
        return 21;
      else if (currPressure >= 990 and currPressure < 1000)
        return 22;
      else if (currPressure >= 980 and currPressure < 990)
        return 24;
      else if (currPressure >= 970 and currPressure < 980)
        return 26;
      else if (currPressure < 970)
        return 26;
    }
  } else if (currPressure > prevPressure) {
    /* RAISING */
    if (mon >= 4 and mon <= 9) {
      /* SUMMER */
      if (currPressure >= 1030)
        return 1;
      else if (currPressure >= 1020 and currPressure < 1030)
        return 2;
      else if (currPressure >= 1010 and currPressure < 1020)
        return 3;
      else if (currPressure >= 1000 and currPressure < 1010)
        return 7;
      else if (currPressure >= 990 and currPressure < 1000)
        return 9;
      else if (currPressure >= 980 and currPressure < 990)
        return 12;
      else if (currPressure >= 970 and currPressure < 980)
        return 17;
      else if (currPressure < 970)
        return 17;
    } else {
      /* WINTER */
      if (currPressure >= 1030)
        return 1;
      else if (currPressure >= 1020 and currPressure < 1030)
        return 2;
      else if (currPressure >= 1010 and currPressure < 1020)
        return 6;
      else if (currPressure >= 1000 and currPressure < 1010)
        return 7;
      else if (currPressure >= 990 and currPressure < 1000)
        return 10;
      else if (currPressure >= 980 and currPressure < 990)
        return 13;
      else if (currPressure >= 970 and currPressure < 980)
        return 17;
      else if (currPressure < 970)
        return 17;
    }
  } else {
    if (currPressure >= 1030)
      return 1;
    else if (currPressure >= 1020 and currPressure < 1030)
      return 2;
    else if (currPressure >= 1010 and currPressure < 1020)
      return 11;
    else if (currPressure >= 1000 and currPressure < 1010)
      return 14;
    else if (currPressure >= 990 and currPressure < 1000)
      return 19;
    else if (currPressure >= 980 and currPressure < 990)
      return 23;
    else if (currPressure >= 970 and currPressure < 980)
      return 24;
    else if (currPressure < 970)
      return 26;
  }
}

String interpretWeather(int Z, float seaPressure) {
  if (pressureArray[9] > 0 and pressureArray[0] > 0) {
    if (pressureArray[9] + pressureArray[8] + pressureArray[7] - pressureArray[0] - pressureArray[1] -
        pressureArray[2] >= 3) {
      printSerial("RAISING");

      if (Z < 3)
        return "Sunny";
      else if (Z >= 3 and Z <= 9)
        return "Overcast";
      else if (Z > 9 and Z <= 17)
        return "Cloudy";
      else if (Z > 17)
        return "Rainy";

    } else if (pressureArray[0] + pressureArray[1] + pressureArray[2] - pressureArray[9] - pressureArray[8] -
               pressureArray[7] >= 3) {
      printSerial("FALLING");

      if (Z < 4)
        return "Sunny";
      else if (Z >= 4 and Z < 14)
        return "Overcast";
      else if (Z >= 14 and Z < 19)
        return "Worsening";
      else if (Z >= 19 and Z < 21)
        return "Cloudy";
      else if (Z >= 21)
        return "Rainy";

    } else {
      printSerial("STEADY");

      if (Z < 5)
        return "Sunny";
      else if (Z >= 5 and Z <= 11)
        return "Overcast";
      else if (Z > 11 and Z < 14)
        return "Cloudy";
      else if (Z >= 14 and Z < 19)
        return "Worsening";
      else if (Z > 19) {
        return "Rainy";
      }
    }
  } else {
    printSerial("Still computing");

    if (seaPressure < 1005)
      return "Rainy";
    else if (seaPressure >= 1005 and seaPressure <= 1015)
      return "Cloudy";
    else if (seaPressure > 1015 and seaPressure < 1025)
      return "Overcast";
    else
      return "Rainy";
  }
}

void printSerial(String s) {
  Serial.print(s);
  Serial.println();
}
